MODULE likelihood

  USE GLOBVAR
  USE RANDOM
  
  IMPLICIT NONE

CONTAINS

SUBROUTINE frequency(ii)
  
  ! This subroutine calculates transition probabilities, steady-state distribution, and etc
  ! All of these are done based on stored data called "dta"
  
  IMPLICIT NONE
  
  INTEGER, INTENT(IN) :: ii
  INTEGER :: m,s,ss,t,s1,s2
  REAL(8) :: temp,PP(4,4)
    
  ! Calculate Markov transition probability named "sigma" and market level prob named "sigma_j"
  sigma_j=0.0d0
  DO m=1,N_mar
      DO s1=1,4
          DO s2=1,4
              s=0
              ss=0
              DO t=1,N_time
                  IF(dta(m,t-1)==s1) s=s+1
                  IF(dta(m,t-1)==s1 .AND. dta(m,t)==s2) ss=ss+1
              END DO
              IF(s>0) sigma_j(m,s1,s2) = DBLE(ss)/DBLE(s)
          END DO
      END DO
  END DO
  sigma=0.0d0    
  DO s1=1,4
      DO s2=1,4
          s=0
          ss=0
          DO m=1,N_mar
              DO t=1,N_time
                  IF(dta(m,t-1)==s1) s=s+1
                  IF(dta(m,t-1)==s1 .AND. dta(m,t)==s2) ss=ss+1
              END DO
          END DO
          IF(s>0) sigma(s1,s2) = DBLE(ss)/DBLE(s)
      END DO
  END DO

  ! Calculate fj_as
  fj_as = dta_ccp(:,:,:)
  
  ! Calculate fj_s
  DO m=1,N_mar
      DO s1=1,4
          fj_s(m,s1) = SUM(fj_as(m,s1,:))
      END DO
  END DO
  
  ! Calculate steady-state distribution Qs and market-level state frequency Qs_j
  Qs_j=0.0d0
  DO m=1,N_mar
      DO t=0,N_time
          Qs_j(m,dta(m,t)) = Qs_j(m,dta(m,t)) + 1.0d0
      END DO
      Qs_j(m,:) = Qs_j(m,:)/DBLE(N_time+1)
  END DO
  DO s=1,4
      Qs(s) = SUM(Qs_j(:,s))/DBLE(N_mar)
  END DO
  
  ! Calculate Qsst and iotaP for Ts(s',T) and Ts(s')  
  Qsst = 0.0d0
  DO s1=1,4
      DO s2=1,4
          DO t=1,N_time
              s = COUNT(dta(:,0)==s1 .AND. dta(:,t)==s2)
              ss =COUNT(dta(:,0)==s1)
              IF(ss>0) Qsst(s1,s2,t) = DBLE(s)/DBLE(ss)
          END DO
      END DO
  END DO  
  iotaP = 0.0d0
  DO t=1,N_time
      PP=sigma
      DO s=2,t
          PP = MATMUL(PP,sigma)
      END DO
      iotaP(:,:,t) = PP
  END DO
    
END SUBROUTINE frequency


FUNCTION test_stat(ii)
  
  ! Calculate 4 test statistics
  ! 1. CCP test based on sigma_j and sigma
  ! 2. CCP test based on sigma_j and sigma
  ! 3. T_Q, SSC test based on Qs_j and Qs
  ! 4. Ts(s')
  
  IMPLICIT NONE
  
  INTEGER, INTENT(IN) :: ii
  INTEGER :: i,t,k,j
  REAL(8) :: temp,temp1,test_stat(N_stat),temp2
  INTEGER :: m,s,s1,s2
  
  ! 1. Conditional choice probability test
  sigma_ccp = 0.0d0
  DO s1=1,4
      DO s2=1,4
          IF(SUM(fj_s(:,s1))>0) sigma_ccp(s1,s2) = DBLE(SUM(fj_as(:,s1,s2)))/DBLE(SUM(fj_s(:,s1)))
      END DO
  END DO
  temp1=0.0d0
  DO m=1,N_mar
      DO s1=1,4
          DO s2=1,4
              temp2 = sigma_ccp(s1,s2)*DBLE(fj_s(m,s1))
              temp = DBLE(fj_as(m,s1,s2)) - temp2
              IF(temp2>0.0d0) temp1 = temp1 + temp*temp/temp2
          END DO
      END DO
  END DO
  test_stat(1) = temp1

  ! 2. Optimal conditional choice probability test
  temp1=0.0d0
  DO m=1,N_mar
      DO s1=1,4
          DO s2=1,4                    
              IF(sigma(s1,s2)>0.0d0 .AND. sigma_j(m,s1,s2)>0.0d0) THEN
                  temp = LOG(sigma_j(m,s1,s2)/sigma(s1,s2))
                  temp1 = temp1 + DBLE(fj_s(m,s1))*sigma_j(m,s1,s2)*temp
              END IF
          END DO
      END DO
  END DO
  test_stat(2) = temp1*2.0d0
  
  ! 3. T_Q, Steady-state distribution test without denominator
  temp1=0.0d0
  DO m=1,N_mar
      DO s=1,4
          temp = Qs_j(m,s) - Qs(s)
          temp1 = temp1 + temp*temp
      END DO
  END DO
  test_stat(3) = temp1*DBLE(N_time)
    
  ! 4. Ts(s')
  temp1=0.0d0
  DO s1=1,4
      DO s2=2,4
          temp = SUM(Qsst(s1,s2,:)) - SUM(iotaP(s1,s2,:))
          temp1 = temp1 + temp*temp
      END DO
  END DO
  test_stat(4) = temp1
    
END FUNCTION test_stat


SUBROUTINE boot(sigma_local)
  
  ! Make a parametric bootstrap sample by simulation
    
  IMPLICIT NONE
  
  REAL(8), INTENT(IN) :: sigma_local(4,4)
  INTEGER :: m,t,s1,s2,s,cc
  REAL(8) :: temp,u
  
  ! Make bootstrap sample
  IF(initial==1) THEN
      DO m=1,N_mar
          s=1
          DO t=1,Sno+N_time
              u = Sample_Uniform(0.0d0,1.0d0)
              temp=0.0d0
              DO s1=1,4
                  temp=temp+sigma_local(s,s1)
                  IF(u<temp) THEN
                      IF(t>=Sno) THEN
                          dta(m,t-Sno) = s1
                      END IF
                      s=s1
                      EXIT
                  END IF
              END DO
          END DO
      END DO
  ELSE
      DO m=1,N_mar
          s=dta(m,0)
          DO t=1,N_time
              u = Sample_Uniform(0.0d0,1.0d0)
              temp=0.0d0
              DO s1=1,4
                  temp=temp+sigma_local(s,s1)
                  IF(u<temp) THEN
                      dta(m,t) = s1
                      s=s1
                      EXIT
                  END IF
              END DO
          END DO
      END DO
  END IF
  
  DO m=1,N_mar
      DO s1=1,4
          DO s2=1,4
              cc=0
              DO t=1,N_time
                  IF(dta(m,t-1)==s1 .AND. dta(m,t)==s2) cc=cc+1
              END DO
              dta_ccp(m,s1,s2) = DBLE(cc)
          END DO
      END DO
  END DO

END SUBROUTINE boot

END MODULE likelihood

